import numpy as np

def _check_nan_padding(a: np.ndarray, b: np.ndarray) -> bool:
    # assert time series are only NaN-padded at the beginning and the end, there are no NaN values in between
    test = lambda ts, start_nan, end_nan: np.count_nonzero(np.diff(np.isnan(ts))) == int(start_nan) + int(end_nan)
    assert len(a) > 0, "Input array is empty"
    assert len(b) > 0, "Input array is empty"
    return test(a, np.isnan(a[0]), np.isnan(a[-1])) and test(b, np.isnan(b[0]), np.isnan(b[-1]))


def _remove_nan_padding_single(a: np.ndarray) -> np.ndarray:
    """ Remove all NaN values from series"""
    a = a[np.logical_not(np.isnan(a))]
    return a

def _remove_nan_padding(a: np.ndarray, b: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """ Remove NaN values from two time series"""
    a = _remove_nan_padding_single(a)
    b = _remove_nan_padding_single(b)
    return a, b


def _var_nan_sum(a: np.ndarray, axis: int = 0, alpha: float = 0.0) -> np.ndarray:
    """ Calculate the sum of a matrix along an axis, but allow alpha percentage of NaN values.

    At indices where the proportion is NaN values is less than alpha, perform the sum as usual.
    At indices where the proportion is NaN values is greater than alpha, return NaN.

    :param a: input matrix
    :param axis: axis along which to sum
    :param alpha: proportion of NaN values allowed
    :return: sum of `a` along `axis`
    """

    assert 0.0 <= alpha <= 1.0
    assert 0 <= axis < len(a.shape)

    # Calculate the number of NaN values along the axis
    nans = np.isnan(a)
    nans = np.sum(nans, axis=axis)

    # Calculate the proportion of NaN values along the axis
    nans = np.divide(nans, a.shape[axis])

    # Calculate the sum of `a` along `axis`
    s = np.nansum(a, axis=axis)

    # Replace the sum with NaN where the proportion of NaN values is greater than alpha
    s[nans > alpha] = np.nan

    return s


import numpy as np
import warnings
from matplotlib import pyplot as plt

from .time_series_alignment import sync_n_series, get_score_and_shift


def single_iteration(prototype: np.ndarray, class_segments: list[np.ndarray], learning_rate: float,
                     previous_t: np.ndarray, prototyping_function=None,
                     exclusion_zone: float = 0.0) -> tuple[
    np.ndarray, np.ndarray]:
    """ Perform a single iteration of prototype generation

    Given the previous prototype, the class segments and the learning rate, calculate the new prototype and the new
    time indices for the prototype.

    The new prototype is calculated using the prototyping_function, which must be specified. The prototyping_function
    must have the following signature:
    previous_prototype: np.ndarray, synced_series: list[np.ndarray], similarities: np.ndarray, learning_rate: float -> np.ndarray

    :param prototype: Previous prototype
    :param class_segments: List of time series in the class
    :param learning_rate: Learning rate
    :param previous_t: Previous time indices for the prototype
    :param use_diff_for_lags: If True, diff(signal) will be used to determine the lag
    :param prototyping_function: Function to be used to generate the new prototype
    :return: [0] -> new prototype; [1] -> new time indices for the prototype

    """
    scores_and_shifts: list[tuple[float, int]] = [
        get_score_and_shift(prototype, segment,
                            exclusion_zone=exclusion_zone, apply_zero_mean=False) for segment in class_segments]
    scores: list[float] = [x[0] for x in scores_and_shifts]
    shifts: list[int] = [x[1] for x in scores_and_shifts]

    temp = [prototype]
    temp.extend(class_segments)
    ts = [0]
    ts.extend(shifts)
    t_synced, synced_series = sync_n_series(temp, ts)

    # Check for validity of the sync
    # Synced series are stacked into a matrix, and then the sum of each column is calculated.
    # If all columns sum up to NaN, this means thath the next prototype will be all NaN, and the algorithm should stop.
    if np.count_nonzero(np.isnan(np.sum(np.vstack(synced_series), axis=0))) == len(t_synced):
        return prototype, previous_t

    # Transform scores into similarities
    # If any scorr score is infinite (can occur when using eral mode), it is replaced with a large value
    similarities = np.array(scores).reshape(-1, 1)
    max_not_inf = np.max(similarities[np.isfinite(similarities)])
    if max_not_inf <= 0:
        max_not_inf = 1
    similarities = np.nan_to_num(similarities, posinf=max_not_inf * 3)

    # Call to prototyping function to obtain the new prototype
    new_prototype: np.ndarray = prototyping_function(previous_prototype=synced_series[0],
                                                     synced_series=synced_series[1:],
                                                     similarities=similarities,
                                                     learning_rate=learning_rate).reshape(-1)
    return new_prototype, t_synced


def obtain_prototype(class_segments: list[np.ndarray],
                     initial_prototype: np.ndarray | int | None = None,
                     max_iterations: int = 10,
                     learning_rate: float = 0.3,
                     min_length: int = -1,
                     prototyping_function=None,
                     exclusion_zone: float = 0.0,
                     min_rmse: float|None = 0.0) -> np.ndarray | None:
    """ Generate prototype from class segments

    Perform several iterations of prototype generation. In each iteration, the prototype is generated using the
    prototyping_function, which must be specified. The prototyping_function must have the following signature:
    previous_prototype: np.ndarray, synced_series: list[np.ndarray], similarities: np.ndarray, learning_rate: float -> np.ndarray

    The learning rate is constant across all iterations. It defines the rate at which the prototype is updated and moved
    towards the training samples.

    This function supports plotting, if ax_series and ax_final are specified.

    :param class_segments: List of time series in the class
    :param initial_prototype: Initial prototype. If None, a random segment will be selected from the class segments
    :param max_iterations: Maximum number of iterations
    :param learning_rate: Learning rate
    :param use_diff_for_lags: If True, diff(signal) will be used to determine the lag
    :param min_length: Minimum length of the prototype. If the prototype is shorter than this, the algorithm will stop
    :param prototyping_function: Function to be used to generate the new prototype. See `alignment_prototyping_functions.py` for examples`
    :param use_eral: If True, the eral mode will be used
    :param exclusion_zone: Exclusion zone
    :param min_rmse: Minimum RMSE. If the RMSE of prototype change is lower than this, the algorithm will stop. If None, will use 1% of data range.
    :return: The final prototype
    """

    assert type(class_segments) is list
    if len(class_segments) == 0:
        raise Exception("No class segments provided")
    if len(class_segments) == 1:
        return class_segments[0]

    if initial_prototype is None:
        all_sample_indices = list(range(len(class_segments)))
        while len(all_sample_indices) > 0:
            initial_prototype_idx = all_sample_indices[0]
            all_sample_indices.pop(0)
            prototype = obtain_prototype(class_segments, initial_prototype=initial_prototype_idx,
                                         max_iterations=max_iterations, learning_rate=learning_rate,
                                         min_length=min_length,
                                         prototyping_function=prototyping_function,
                                         exclusion_zone=exclusion_zone,
                                         min_rmse=min_rmse)
            if prototype is not None:
                return prototype
        warnings.warn("Could not find a prototype, returning the first segment")
        return class_segments[0]

    elif type(initial_prototype) is int:
        prototype = class_segments[initial_prototype]
    else:
        prototype = initial_prototype


    if prototyping_function is None:
        raise Exception("prototyping_function must be specified, it must be a function with the following signature: "
                        "previous_prototype: np.ndarray, synced_series: list[np.ndarray], similarities: np.ndarray, "
                        "learning_rate: float -> np.ndarray. "
                        "See `alignment_prototyping_functions.py` "
                        "for examples")


    if min_rmse is None:
        data_range = np.max([np.max(segment) for segment in class_segments]) - np.min([np.min(segment) for segment in class_segments])
        min_rmse = 0.01 * data_range

    t_synced = np.arange(0, len(prototype))
    for i in range(max_iterations):
        new_prototype, t_synced = single_iteration(prototype, class_segments, learning_rate, t_synced,
                                                   prototyping_function=prototyping_function,
                                                   exclusion_zone=exclusion_zone)

        if np.count_nonzero(np.isnan(new_prototype)) == len(new_prototype):
            raise Exception("Prototype is all nan!")

        if min_length > 0 and np.count_nonzero(np.logical_not(np.isnan(new_prototype))) < min_length:
            if i == 0:
                return None
            else:
                break

        # Todo: Add a stopping criterion based on the change in the prototype
        if len(prototype) == len(new_prototype):
            if np.all(prototype == new_prototype):
                break

        if len(prototype) == len(new_prototype):
            rmse = np.sqrt(np.nanmean((prototype - new_prototype) ** 2))
            if rmse < min_rmse:
                break
        else:
            pass


        prototype = new_prototype
    return prototype




import numpy as np
from .time_series_helpers import _check_nan_padding, _remove_nan_padding
from .eral_score import eral_score
from scipy.signal import correlate, correlation_lags


def get_score_and_shift(in1: np.ndarray, in2: np.ndarray, apply_zero_mean: bool = True, exclusion_zone: float = 0.0) -> tuple[float, int]:
    """ Helper function returning the eral score and calculated lag for two time series

    :param in1: first time series (passed to `scipy.signal.correlate` as first argument). May include NaNs.
    :param in2: second time series (passed to `scipy.signal.correlate` as second argument). May include NaNs.
    :param apply_zero_mean: If `True`, zero mean will be applied to both time series
    :param exclusion_zone: Proportion of forbidden shifts. Eg. 0.1 means that the optimal shift can not be in the first or last 10% of the time series.
    :return: First output: eral score (the highest value accross all lags); Second argument: optimal lag (at highest eral score)
    """

    if not 0 <= exclusion_zone < 0.5:
        raise ValueError("exclusion_zone must be between 0 and 0.5")


    in1_diff = in2_diff = None
    corr_diff = lags_diff = None

    # region Handle NaNs
    _check_nan_padding(in1, in2)
    in1_nan_left_padding_count = in2_nan_left_padding_count = 0
    for x in in1:
        if np.isnan(x):
            in1_nan_left_padding_count += 1
        else:
            break
    for x in in2:
        if np.isnan(x):
            in2_nan_left_padding_count += 1
        else:
            break
    in1, in2 = _remove_nan_padding(in1, in2)
    # endregion

    # Apply zero mean
    if apply_zero_mean:
        # check if in1 has nans
        if np.isnan(in1).any():
            in1 = in1 - np.nanmean(in1)
        else:
            in1 = in1 - np.mean(in1)

        # check if in2 has nans
        if np.isnan(in2).any():
            in2 = in2 - np.nanmean(in2)
        else:
            in2 = in2 - np.mean(in2)

    _eral = eral_score(in1, in2)
    old_settings = None
    if np.min(_eral) == 0:
        old_settings = np.seterr(divide='ignore')  # _eral can be 0 if the two signals are identical.
        # This would cause a division by zero warning, which is not necessary.
    corr = 1 / _eral
    if old_settings is not None:
        np.seterr(**old_settings)

    lags = correlation_lags(len(in1), len(in2), 'full')  # eral uses the same lags as 'full'
    assert len(corr) == len(lags)

    if exclusion_zone > 0:
        # Exclude the forbidden shifts
        exclusion_zone = int(len(corr) * exclusion_zone)
        corr[:exclusion_zone] = 0
        corr[-exclusion_zone:] = 0

    # We must account for the left padding of NaN values when returning the lag
    argmax = np.argmax(corr)
    return corr[argmax], lags[argmax] + in1_nan_left_padding_count - in2_nan_left_padding_count
    # return np.max(corr), lags[np.argmax(corr)] # + in1_nan_left_padding_count - in2_nan_left_padding_count


def _get_optimal_shift_for_two_series(in1: np.ndarray, in2: np.ndarray):
    """ Helper function returning the optimal shift for two time series

    Does not use circular padding, and does not use diff for lag calculation

    :param in1: first time series (passed to `scipy.signal.correlate` as first argument)
    :param in2: second time series (passed to `scipy.signal.correlate` as second argument)
    :param mode: mode of signal alignment (`full`, `valid`, `eral`)
    :return: optimal shift
    """
    return get_score_and_shift(in1, in2)[1]


def sync_2_series(in1: np.ndarray, in2: np.ndarray, shift: int | None = None) -> tuple[
    np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """ Helper function to synchronize two time series by shifting the second one

    :param in1: The 'background' time series - is stationary
    :param in2: The 'foreground' time series - is shifted according to the `shift` parameter
    :param shift: The number of samples `in2` should be shifted. Positive = shift 'right'; Negative = shift 'left'. If `None`, the optimal shift will be calculated.
    :param mode: `mode` used in aligning the time series (`full`, `valid`, `eral`)
    :return: [0] -> new indices for in1;
             [1] -> shifted in1;
             [2] -> new indices for in2;
             [3] -> shifter in2
    """

    if shift is None:
        shift = _get_optimal_shift_for_two_series(in1, in2)

    signal_2_shifted = np.ones(np.abs(shift) + len(in2)) * np.nan
    signal_1_shifted = np.ones(np.abs(shift) + len(in1)) * np.nan
    if shift < 0:
        shifted_indices_1 = range(-np.abs(shift), len(in1))
        shifted_indices_2 = range(-np.abs(shift), len(in2))
        signal_2_shifted[0:len(in2)] = in2
        signal_1_shifted[np.abs(shift):] = in1
    else:
        shifted_indices_1 = range(0, len(signal_1_shifted))
        shifted_indices_2 = range(0, len(signal_2_shifted))
        signal_2_shifted[np.abs(shift):] = in2
        signal_1_shifted[0:len(in1)] = in1

    return shifted_indices_1, signal_1_shifted, shifted_indices_2, signal_2_shifted


def sync_n_series(series: list[np.ndarray], shifts: list[int]) -> tuple[np.ndarray, list[np.ndarray]]:
    """ Shift all series in `series` by `shifts` samples

    :param series: A list of time series to be shifted
    :param shifts: The shifts (a list of integers)
    :return: [0] -> new common indices for all series;
             [1] -> list shifted series. np.nan where series has no value.
    """
    output_series: list[np.ndarray] = []

    time_start: int = np.min(shifts)
    time_end: int = np.max(np.array(shifts) + np.array(list(map(lambda x: len(x), series)))).__int__()
    output_time: np.ndarray = np.array(list(range(time_start, time_end)))

    for i in range(len(shifts)):
        ser: np.ndarray = series[i]
        shift: int = shifts[i]
        shifted: np.ndarray = np.ones(len(output_time)) * np.nan
        shifted[shift - time_start:shift - time_start + len(ser)] = ser
        output_series.append(shifted)

    return output_time, output_series


def sync_n_series_to_prototype(prototype: np.ndarray, series: list[np.ndarray], shifts: list[int] = None, exclusion_zone: float = 0.0) -> tuple[
    np.ndarray, list[np.ndarray], np.ndarray]:
    """ Shift all series in `series` by `shifts` samples and align them to the prototype

    :param prototype: The prototype to which all series should be aligned
    :param series: A list of time series to be shifted
    :param shifts: The shifts (a list of integers). If `None`, the optimal shifts will be calculated.
    :return: [0] -> new common indices for all series;
             [1] -> list shifted series. np.nan where series has no value.
             [2] -> shifted prototype
    """

    if shifts is None:
        shifts = [get_score_and_shift(prototype, ser, exclusion_zone=exclusion_zone)[1] for ser in series]

    # We will use sync_n_series which takes a list of time series and allignes them according to the shifts.
    # We pass the prototype as the first series in the list, and add it the shift of 0.
    temp_data = [prototype]
    temp_data.extend(series)
    temp_shifts = [0]
    temp_shifts.extend(shifts)

    synced_time, synced_data = sync_n_series(series=temp_data, shifts=temp_shifts)

    # First synced_data is the prototype
    return synced_time, synced_data[1:], synced_data[0]
